try:
    from pyspark.sql import SparkSession
    from pyspark.sql.dataframe import DataFrame
    from pyspark.sql import functions as F
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")


__all__ = [
    "get_dwd_data",
    "get_operator_data",
]


def get_dwd_data(
        db_name: str,
        table_name: str,
        column: str,
        start_date: str,
        end_date: str,
        spark: SparkSession,
) -> DataFrame:
    """
    清洗层取数函数

    :param db_name: 数据库名称
    :param table_name: 表名
    :param column: 取数字段名称
    :param start_date: 开始时间 yyyyMMdd
    :param end_date: 结束时间 yyyyMMdd
    :param spark: 数据库连接
    :return: 取数结果
    """
    query = f"""
        SELECT uuid, {column}, load_date
        FROM {db_name}.{table_name}
        WHERE load_date BETWEEN '{start_date}' AND '{end_date}'
        """
    df = spark.sql(query)
    return df


def get_operator_data(
        df: DataFrame,
        column: str,
        start_date: str,
        end_date: str,
) -> DataFrame:
    """
    清洗层取数函数
    :param df: 原始数据
    :param column: 取数字段名称
    :param start_date: 开始时间 yyyyMMdd
    :param end_date: 结束时间 yyyyMMdd
    :return: 取数结果
    """
    df = df.where(F.col(start_date).between(start_date, end_date))
    return df.select("uuid", column, "load_date")



import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from scipy.stats import spearmanr
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = [
    "PearsonBins"
]


class PearsonBins(WOEMixin):
    bins: Optional[List[Union[float, int]]]

    def __init__(
            self,
            variables: Series,
            target: Series,
            r: Optional[float] = 0.8,
            min_bin: Optional[int] = 3,
            max_bin: Optional[int] = 20,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            fill_bin: Optional[bool] = True,
            decimal: Optional[int] = 6,
            weight: Optional[Any] = None,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：使用 Pearson 相关系数方法进行分箱分析。

        :param r: 相关系数阈值
        :param min_bin: 最小分箱数
        :param max_bin: 最大分箱数

        Example:
            woe = PearsonBins(r=1, min_bin=3, max_bin=20)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.r = r
        self.min_bin = min_bin
        self.max_bin = max_bin
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target, weight=weight)

    def pearson_bins(self):
        """"""
        n = self.max_bin
        r = 0
        while np.abs(r) < self.r and n >= self.min_bin:
            try:
                _bucket, self.bins = pd.qcut(self.df_data.variables, n, retbins=True, duplicates='drop')
                bucket_group = pd.DataFrame({
                    "variables": self.df_data.variables,
                    "target": self.df_data.target,
                    "bucket": _bucket,
                }).groupby('bucket', as_index=True, observed=False)
                r, p = spearmanr(bucket_group.mean().variables, bucket_group.mean().target)
            except Exception as e:
                self._logger(msg=f"[{__class__.__name__}] Error: {e}.", color="red")
            n -= 1
        self.bins[0] = -np.inf
        self.bins[-1] = np.inf

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """
        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        self.pearson_bins()
        _bucket, _bins = pd.cut(
            self.df_data.variables, self.bins, include_lowest=True, retbins=True, duplicates='drop')
        bucket = pd.DataFrame({
            'variables': self.df_data.variables,
            'target': self.df_data.target,
            'bucket': _bucket,
        }).groupby('bucket', as_index=True, observed=False)
        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe



import numpy as np
import pandas as pd
from typing import Optional, Literal, Tuple, Union, List, Any
from pandas.api.types import is_numeric_dtype
from kivi.evaluate.utils import BinsMixin
from kivi.evaluate.BucketsUtils import psi


__all__ = ['PSI']


class PSI(BinsMixin):
    cut_bins: Optional[List[float]]

    def __init__(
            self,
            expected: pd.Series,
            actual: pd.Series,
            bins: Union[int, float, List[Union[int, float]]] = 10,
            cut: Literal["cut", "qcut"] = 'cut',
            _min: Optional[float] = None,
            _max: Optional[float] = None,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            return_columns: Optional[List[str]] = None,
            decimal: Optional[int] = 6,
            **kwargs: Any,
    ):
        """
        计算 PSI
        :param expected: 期望数据
        :param actual: 实际数据
        :param bins: 分箱数目或分箱截断点
        :param cut: 分箱类型 等距`cut` 或 等频`qcut`
        :param _min: 最小值
        :param _max: 最大值
        :param abnormal_vals: 异常值
        :param return_columns: 返回列名
        :param decimal: 保留小数位数
        """
        self.expected = expected
        self.actual = actual
        self.bins = bins
        self.cut = cut
        self._min = _min
        self._max = _max
        self.abnormal_vals = abnormal_vals
        self.return_columns = return_columns
        self.decimal = decimal
        self.cut_bins = None
        self.kwargs = kwargs
        self.set_return_columns()
        self.is_numeric = self._validate_numeric()

    def set_return_columns(self):
        """"""
        if self.return_columns is None:
            self.return_columns = [
                "min_bin", "max_bin", "Expected", "Actual",
                "Expected_percentage", "Actual_percentage",
                "psi_val", "psi",
            ]

    def _validate_numeric(self) -> bool:
        """"""
        if all([is_numeric_dtype(data) for data in [self.expected, self.actual]]):
            return True
        else:
            return False

    def _get_bins(self):
        """"""
        if isinstance(self.bins, list):
            self.cut_bins = self.bins
        else:
            if self.cut == 'qcut':
                _, self.cut_bins = pd.qcut(self.expected, self.bins, retbins=True, duplicates='drop',)
            elif self.cut == 'cut' and self._min is not None and self._max is not None:
                self.cut_bins = np.linspace(self._min, self._max, self.bins + 1).tolist()
            elif self.cut == 'cut':
                _, self.cut_bins = pd.cut(self.expected, self.bins, retbins=True, include_lowest=True, duplicates='drop')
            self.cut_bins[0] = -np.inf
            self.cut_bins[-1] = np.inf

    def _miss_data_cnt(self) -> Optional[pd.DataFrame]:
        """"""
        sum_ex = sum(self.expected.isna())
        sum_ac = sum(self.actual.isna())
        if sum_ex > 0 or sum_ac > 0:
            na_bin = pd.DataFrame({
                'min_bin': [np.nan], 'max_bin': [np.nan], 'Expected': [sum_ex], 'Actual': [sum_ac],
            })
            return na_bin
        return None

    def _abnormal_val_cnt(self, ) -> Optional[pd.DataFrame]:
        """"""
        abnormal_data = []
        if isinstance(self.abnormal_vals, list):
            for val in self.abnormal_vals:
                sum_ex = (self.expected == val).sum()
                sum_ac = (self.actual == val).sum()
                abnormal_data.append(pd.DataFrame({
                    'min_bin': [val], 'max_bin': [val], 'Expected': [sum_ex], 'Actual': [sum_ac],
                }))
            return pd.concat(abnormal_data, axis=0)
        return None

    def prepare_data(self) -> Tuple[pd.Series, pd.Series]:
        """"""
        expected, actual = self.expected.dropna(), self.actual.dropna()
        if isinstance(self.abnormal_vals, list):
            expected = expected[~expected.isin(self.abnormal_vals)]
            actual = actual[~actual.isin(self.abnormal_vals)]
        return expected, actual

    def _normal_val(self) -> Optional[pd.DataFrame]:
        """"""
        expected, actual = self.prepare_data()
        df_ex = pd.DataFrame(pd.cut(expected, self.cut_bins, include_lowest=True).value_counts(sort=False))
        df_ac = pd.DataFrame(pd.cut(actual, self.cut_bins, include_lowest=True).value_counts(sort=False))
        df_psi = pd.concat([df_ex, df_ac], axis=1)
        df_psi.columns = ['Expected', 'Actual']
        df_psi = self.add_bins(df_psi.fillna(0))
        return df_psi

    def add_bins(self, df_psi: pd.DataFrame) -> pd.DataFrame:
        """"""
        df_psi['min_bin'] = self.cut_bins[: -1]
        df_psi['max_bin'] = self.cut_bins[1:]
        return df_psi

    def _value_decimal(self, psi: pd.DataFrame) -> pd.DataFrame:
        """"""
        decimal_columns = [
            "Expected_percentage", "Actual_percentage", "psi_val", "psi"
        ]
        for col in decimal_columns:
             psi[col] = psi[col].round(self.decimal)
        return psi

    def numeric(self):
        """"""
        self._get_bins()
        psi_data = [self._normal_val(), self._miss_data_cnt(), self._abnormal_val_cnt()]
        psi_data = [item for item in psi_data if item is not None]
        df_psi = pd.concat(psi_data, axis=0)
        return psi(df_psi, fill_bin=True)[self.return_columns]

    def _category(self) -> pd.DataFrame:
        """"""
        expected, actual = self.expected.dropna(), self.actual.dropna()
        df_psi = pd.concat([
            pd.DataFrame(expected.value_counts(sort=False)),
            pd.DataFrame(actual.value_counts(sort=False)),
        ], axis=1)
        df_psi.columns = ['Expected', 'Actual']
        df_psi["min_bin"] = df_psi.index
        df_psi["max_bin"] = df_psi.index
        df_psi.sort_index(inplace=True)
        return psi(df_psi, fill_bin=True)[self.return_columns]
    
    def fit(self, ) -> pd.DataFrame:
        """"""
        if self.is_numeric:
            df_psi = self.numeric()
        else:
            df_psi = self._category()
        df_psi = self._value_decimal(df_psi)
        return df_psi





from sklearn.linear_model import RidgeCV
from pandas import DataFrame, Series
from sklearn.metrics import r2_score, mean_squared_error
from sklearn.compose import TransformedTargetRegressor
import numpy as np



class Ridge:

    def __init__(self, data, y_name, alphas, normalize=False, log_y=False):
        """
        Des: Ridge | 岭回归方法
        :param data: 数据集
        :param y_name: y 名称
        :param alphas: 待选alpha参数
        :param normalize: 是否正则化，默认False
        :param log_y: 是否y取对数，默认False
        """
        y = data.pop(y_name)
        self.exog = data.copy()
        self.endog = y.copy()

        self.model = self.ridge(self.exog, self.endog, alphas, normalize, log_y)

        if log_y:
            self.coef = Series(self.model.regressor_.coef_, index=self.exog.columns).abs().sort_values(ascending=False)
            self.intercept = self.model.regressor_.intercept_
            self.alpha = self.model.regressor_.alpha_
        else:
            self.coef = Series(self.model.coef_, index=self.exog.columns).abs().sort_values(ascending=False)
            self.intercept = self.model.intercept_

        y_pre = self.model.predict(self.exog)
        self.mse = mean_squared_error(self.endog, y_pre)
        self.r2_score = r2_score(self.endog, y_pre)


    def ridge(self, X, y, alphas, normalize, log_y):
        """
        Des: Ridge 回归
        :param X:
        :param y:
        :param alphas:
        :param normalize:
        :param log_y:
        :return:
        """
        if log_y:
            _model = RidgeCV(alphas=alphas, normalize=normalize)
            tt_model = TransformedTargetRegressor(
                regressor=_model,
                func=np.log,
                inverse_func=np.exp
            )
            return tt_model.fit(X, y)
        else:
            return RidgeCV(alphas=alphas, normalize=normalize).fit(X, y)






from sklearn.tree import DecisionTreeRegressor, export_graphviz
from sklearn.metrics import mean_squared_error

class Tree:

    def __init__(self, data, y_name):
        """

        :param data:
        :param y_name:
        """

        y = data.pop(y_name)
        self.exog = data.copy()
        self.endog = y.copy()
        self.cols_name = self.exog.columns.to_list()


    def fit(
            self,
            max_depth=2,
            min_samples_split=500,
            min_samples_leaf=50,
    ):
        """

        :param max_depth:
        :param min_samples_split:
        :param min_samples_leaf:
        :return:
        """

        self.model = DecisionTreeRegressor(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
        ).fit(self.exog, self.endog)


    def metrics(self, ):
        """

        :return:
        """
        self.bad_rate = self.endog.sum() / self.endog.count()
        self.mse = mean_squared_error(self.endog, self.model.predict(self.exog))


    def plot_tree(self, class_names=['no', 'yes']):
        """
        Des:
        :param feature_names:
        :param class_names:
        :return:
        """
        import pydot
        from IPython.display import Image
        from sklearn.externals.six import StringIO

        dot_data = StringIO()

        export_graphviz(
            self.model,
            out_file=dot_data,
            feature_names=self.cols_name,
            class_names=class_names,
            filled=True,
            rounded=True,
            special_characters=True,
            proportion=True,
        )

        (graph,) = pydot.graph_from_dot_data(dot_data.getvalue())
        return Image(graph.create_png())




import calendar
from datetime import datetime
from typing import List, Tuple, Union
import pandas as pd
from pandas.tseries.offsets import MonthEnd


__all__ = [
    "get_month_last_day",
    "get_start_and_end_date",
    "get_start_and_end_date_periods",
]


def get_month_last_day(date: datetime) -> datetime:
    """"""
    year = date.year
    month = date.month
    _, last_day = calendar.monthrange(year, month)
    month_last_day = datetime(year, month, last_day)
    return month_last_day


def get_start_and_end_date(
        biz_date: str,
        month: Union[int, str],
) -> str:
    """
    desc: 时间回溯函数
    :param biz_date:
    :param month: 时间周期
    :return: 两个 list 指标开始时间，指标结束时间

    >>> get_start_and_end_date(biz_date='20191010', month=month)
    """
    if isinstance(month, str):
        return '20010131'
    else:
        biz_date = pd.to_datetime(biz_date)
        start_date = biz_date - pd.DateOffset(months=month - 1) + MonthEnd(0)
        return start_date.strftime("%Y%m%d")


def get_start_and_end_date_periods(
        start_date: str,
        end_date: str,
        months: int = 3,
        is_hist: bool = False,
) -> Tuple[List[str], List[str]]:
    """
    desc: 时间回溯函数
    :param is_hist: 是否索引历史全量数据
    :param start_date: 开始时间
    :param end_date: 结束时间，回溯月份
    :param months: 时间周期
    :return: 两个 list 指标开始时间，指标结束时间

    >>> start_dates, end_dates = get_start_and_end_date(start_date='20191010', end_date='20201010', months=3)
    >>> print(start_dates, "\\n", end_dates)
    """
    start_dates = pd.date_range(start=start_date, end=end_date, freq='M')
    if is_hist:
        start_dates = start_dates.astype(str).str.replace('-', '').tolist()
        return ['20101231']*len(start_dates), start_dates
    else:
        end_dates = start_dates - pd.DateOffset(months=months - 1) + MonthEnd(0)
        start_dates = start_dates.astype(str).str.replace('-', '').tolist()
        end_dates = end_dates.astype(str).str.replace('-', '').tolist()
        return end_dates, start_dates
